/**
    @author Reuben Shea
    @date   01-06-18

    The BIC container contains the BIC code, IO ports, registers,
    and the current execution byte location.
 **/

#ifndef BIC_CONTAINER_H
#define BIC_CONTAINER_H
#include <cstdint>
#include <string>
#include <vector>
using std::string;

    // Port_Type is used to define the data type of our registers io ports
enum    Port_Type {PT_float, PT_int, PT_ptr, PT_undf};

    // The port represents a register / io port.
    // It contains the data, the length of the data, and the type of data
struct  port_t
{
    private:
        void*   value = nullptr;
        void setValue(void* v)
        {
            if (value != nullptr)
            {
                if (type == PT_float)
                    delete (double*) value;

                if (type == PT_int)
                    delete (int*) value;

                if (type == PT_ptr)
                    delete [] (uint8_t*) value;
            }
            type = PT_undf;
            value = v;
        }

    public:
        long length = 0;        // In bytes
        Port_Type   type = PT_undf;

        void operator=(double* p){setValue((void*) p); type = PT_float; length = 8;}
        void operator=(int* p){setValue((void*) p); type = PT_int; length = 4;}
        void operator=(void* p){setValue(p); type = PT_ptr;}
        void* getValue(){return value;};
            // Makes a deep-copy as needed:
        port_t  mkCopy(){port_t port;
                         if (type == PT_ptr)
                         {
                             uint8_t* a = new uint8_t[length];
                             for (long i = 0; i < length; ++i)
                                 a[i] = ((uint8_t*)value)[i];

                             port = (void*) a;
                             port.length = length;
                         }
                         if (type == PT_float)
                             port = new double (*((double*) getValue()));
                         if (type == PT_int)
                             port = new int (*((int*) getValue()));
                         return port;
                         };

};

class BIC_Container
{
    public:
        BIC_Container(void* buffer, double buffsize);
        virtual ~BIC_Container();
        uint8_t     getIOCount(){return ioCount;};
        uint8_t     getRegCount(){return regCount;};
        void        getIOTypes(Port_Type * a){for (uint8_t i = 0; i < ioCount; ++i) a[i] = io[i].type;};
        double      getIOSize(uint8_t io){return (double) this -> io[io].length;};
        port_t      getIO(uint8_t id){return io[id];}
        port_t      getReg(uint8_t id){return registers[id];};
        uint16_t    getSeek(){return seek;};
        uint16_t    getBufSize() {return bufsize;};
        int         getCall() {return call;};

        uint8_t     getBufU8();
        uint16_t    getBufU16();
        int         getBufS32();
        double      getBufF64();
        string      getBufS();

        void        addTraceReg(uint8_t id){regTraceList.push_back(id); tracing = true;};
        void        addTraceIO(uint8_t id){ioTraceList.push_back(id); tracing = true;};
        void        pullTraceReg(uint8_t id);
        void        pullTraceIO(uint8_t id);

        void        setSeek(uint16_t seek){this -> seek = seek;};
        void        setIO(uint8_t io, port_t port){ioPrntTrace(io, port); this -> io[io] = port;};
        void        setReg(uint8_t reg, port_t port){regPrntTrace(reg, port); this -> registers[reg] = port;};
        void        setCall(int call){this -> call = call;};


    private:
        void*   buffer = nullptr;       // BIC code (shared data w/ GameMaker)
        port_t* registers = nullptr;    // Array of stored registers
        port_t* io = nullptr;           // Array of stored IO ports
        int call = -1;                       // GameMaker script ID to call upon completion (-1 = none)
        uint16_t seek;                  // Current byte location
        uint16_t bufsize;               // Size of the BIC buffer
        uint8_t regCount, ioCount;
        bool    tracing = false;
        std::vector <uint8_t> regTraceList, ioTraceList;

        void regPrntTrace(uint8_t reg, port_t p);
        void ioPrntTrace(uint8_t io, port_t p);
};

#endif // BIC_CONTAINER_H
