#ifndef INTERFACES_H_INCLUDED
#define INTERFACES_H_INCLUDED
#include "properties.h"

/// -- GameMaker interface definition -- /// START
#ifdef BIC_INTERFACE_GAMEMAKER

// -- GameMaker-Specific Defines -- //
#define GMX extern "C"  __declspec (dllexport)

#include <string>
#include <vector>
#include <mutex>
using std::string;
using std::vector;

// -- GameMaker-specific Functions -- //
    // Following functions are for calling GameMaker functions directly:
void    (*GMAsyncTrigger)(int mapID, int eventID);
int     (*GMMapCreate)(int keyCount, ...);
bool    (*GMMapAddDouble)(int mapID, const char* key, double value);
bool    (*GMMapAddString) (int mapID, const char* key, const char* value);
    // Define our GM function pointers:
GMX void RegisterCallbacks(void* function1, void* function2, void* function3, void* function4)
{
    GMAsyncTrigger = (void (*) (int, int))  (function1);
    GMMapCreate = (int (*) (int, ...)) (function2);
    GMMapAddDouble = (bool (*) (int _index, const char*, double)) (function3);
    GMMapAddString = (bool (*) (int _index, const char*, const char*))   (function4);
}
void errorCommit(const char * err);
const int GM_EVENT = 75;     // ID 75 triggers system async event
vector <string>          runner_errors; // Holds all our errors. Redundant, really.
std::mutex gmConnect;   // Must lock threads when sending data to GameMaker or we will have clashes

// -- GameMaker interfacing functions -- //
/*
    @desc   Creates a new container entity and stores the pointer ID for reference.
    @param  {void*}     Data buffer that contains byte code
    @param  {double}    GM script ID to call upon completion of code
    @param  {BIC_Container*}    Pointer that our container can be written to
    @returns    {void}
 */
GMX void  containerCreate(void* buffer, uint32_t* ptr, uint16_t* buffsize)
{
    BIC_Container* container = new BIC_Container(buffer, *buffsize);
    ptr[0] = (uint32_t) container;
}

/*
    @desc   Destroys an existing container.
    @param  {BIC_Container*}    Container to destroy
    @returns    {void}
 */
GMX void    containerDestroy(uint32_t* ptr)
{
    delete (BIC_Container*) (ptr[0]);
}

/*
    @desc   Sets the GameMaker script ID to call when our task finishes.
    @param  {uint32_t*} Pointer to function id
    @returns    {void}
 */
GMX void    containerSetCall(uint32_t* ptr)
{
    BIC_Container* container = (BIC_Container* )ptr[0];
    container -> setCall((int) ptr[1]);
}

/*
    @desc   Sets an IO port value to a double.
    @param  {uint32_t*} Pointer to our BIC_Container to modify
    @param  {double}    IO port to modify
    @param  {double}     Value to set the port to
    @returns    {void}
 */
GMX void    containerSetIOFloat(uint32_t* index, double IOPort, double value)
{
    BIC_Container* container = (BIC_Container*)(index[0]);

    if ((uint8_t) IOPort >= container -> getIOCount())
    {
        errorCommit("Invalid IO Port!");
        return;
    }

    port_t port;
    port = new double (value);

    container -> setIO((uint8_t) IOPort, port);
}

/*
    @desc   Sets an IO port value to an int.
    @param  {uint32_t*} Pointer to our BIC_Container to modify
    @param  {double}    IO port to modify
    @param  {double}     Value to set the port to
    @returns    {void}
 */
GMX void    containerSetIOInt(uint32_t* index, double IOPort, int* value)
{
    BIC_Container* container = (BIC_Container*)(index[0]);

    if ((uint8_t) IOPort >= container -> getIOCount())
    {
        errorCommit("Invalid IO Port!");
        return;
    }

    port_t port;
    port = new int (*value);

    container -> setIO((uint8_t) IOPort, port);
}

/*
    @desc   Sets an IO port value to a pointer
    @param  {uint32_t*} Pointer to our BIC_Container to modify
    @param  {double}    IO port to modify
    @param  {void*}     Value to set the port to
    @param  {double}    Length of the array
    @returns    {void}
 */
 // NOTICE:     the "index" shares a second value of buffer length
 //             because GameMaker cannot pass pointers if the argument
 //             count is > 3
GMX void    containerSetIOPtr(uint32_t* index, double IOPort, void* value)
{
    BIC_Container* container = (BIC_Container*)(index[0]);
    uint32_t len = index[1];

    if ((uint8_t) IOPort >= container -> getIOCount())
    {
        errorCommit("Invalid IO Port!");
        return;
    }

    port_t port;
    uint8_t* ptr = new uint8_t[len];
    for (uint32_t i = 0; i < len; ++i)
        ptr[i] = ((uint8_t*)value)[i];

    port = ptr;
    port.length = len;

    container -> setIO((uint8_t) IOPort, port);
}

/*
    @desc   Returns the number of IO ports for a specific container
    @param  {uint32_t*} pointer to our container
    @returns    {double}    number of ports
 */
GMX double  containerGetIOCount(uint32_t* index)
{
    BIC_Container* container = (BIC_Container*)(index[0]);
    return container -> getIOCount();
}

/*
    @desc   Returns the types for each IO port so we know what to read when we grab
            the values.
    @param  {uint32_t*} Pointer to our BIC_Container to read from
    @param  {Port_Type*}Buffer will be filled with the port types.
    @returns    {void}
 */
GMX void    containerGetIOType(uint32_t* index, uint8_t* data)
{
    BIC_Container* container = (BIC_Container*)(index[0]);
    uint8_t ioCount = container -> getIOCount();
    Port_Type ports[ioCount];

    container -> getIOTypes(ports);
    for (uint8_t i = 0; i < ioCount; ++i)
        data[i] = (uint8_t)ports[i];
}

/*
    @desc   Returns the number of bytes required to store all the IO data.
    @param  {uint32_t*} Pointer to our BIC_Container to read from
    @returns    {double}    Bytes needed for IO data
 */
GMX double   containerGetIOSize(uint32_t* index, double io)
{
    BIC_Container* container = (BIC_Container*)(index[0]);
    return container -> getIOSize(io);
}

/*
    @desc   Returns the data for a specific IO port.
    @param  {uint32_t*} Pointer to our BIC_Container to read from
    @param  {void*}     Buffer to write the data to
    @param  {double}    Port ID to read from
    @returns    {void}
 */
GMX void    containerGetIOData(uint32_t* index, void* data, double io)
{
    BIC_Container* container = (BIC_Container*)(index[0]);
    uint32_t offset = 0;
    uint8_t i = io;

    port_t port = container -> getIO(i);
    if (port.type == PT_float)
    {
        double* offData = (double*)&(((uint8_t*)data)[offset]);
        offData[0] = ((double*) port.getValue())[0];
        offset += 8;
    }

    if (port.type == PT_int)
    {
        int* offData = (int*)&(((uint8_t*)data)[offset]);
        offData[0] = ((int*) port.getValue())[0];
        offset += 4;
    }

    if (port.type == PT_ptr)
    {
        uint8_t* offData = &(((uint8_t*)data)[offset]);
        for (long i = 0; i < port.length; ++i)
            offData[i] = ((uint8_t*) port.getValue())[i];

        offset += port.length;
    }

}

/*
    @desc   Returns the number of bytes required to store all the errors
    @returns    bytes
 */
GMX double  errorsGetBytes()
{
    uint32_t    bytes = 0;
    for (uint32_t i = 0; i < runner_errors.size(); ++i)
        bytes += runner_errors[i].length();

    bytes += runner_errors.size(); // Null termination
    return bytes;
}

/*
    @desc   Returns all the errors as strings in a single buffer
    @param  Buffer to contain the data
    @returns    {void}
 */
GMX void  errorsGetString(char* data)
{
    uint32_t offset = 0;
    for (uint32_t i = 0; i < runner_errors.size(); ++i)
    {
        auto str = runner_errors[i].c_str();
        uint32_t offset_str = 0;
        while (true)
        {
            char c = str[ offset_str];
            ++offset_str;
            data[offset] = c;
            ++offset;

            if (c == 0)
                break;
        }
    }
}

/*
    @desc   Creates a new idle runner.
    @param  {uint32_t*} buffer to write the runner pointer to
    @returns    {void}
 */
GMX void    runnerCreate(uint32_t* ptr)
{
    BIC_Runner* runner = new BIC_Runner();
    ptr[0] = (uint32_t) runner;
}

/*
    @desc   Destroys the specified runner and stops running thread.
    @param  {uint32_t*} pointer to runner
    @returns    {void}
 */
GMX void    runnerDestroy(uint32_t* ptr)
{
    BIC_Runner* runner = (BIC_Runner*) ptr[0];
    runner -> stopThread();
    runner -> joinThread();
    delete runner;
}

/*
    @desc   Starts a thread on the runner to start processing
            all the submitted containers.
    @param  {uint32_t*} pointer of runner to start
    @returns    {double}    whether or not the thread started
 */
GMX double    runnerStart(uint32_t* ptr)
{
    BIC_Runner* runner = (BIC_Runner*) ptr[0];

    if (runner -> getRunning())
    {
        errorCommit("Cannot start thread, thread already running!");
        return 0.;
    }

    if (runner -> getTaskCount() <= 0)
    {
        errorCommit("Cannot start thread, empty tasks!");
        return 0.;
    }

    try
    {
        runner -> startThread();
    }
    catch (...)
    {
        errorCommit("Failed to create thread!");
        return 0.;
    }
    return 1.;
}

/*
    @desc   Determines if the specified runner is currently computing tasks.
    @param  {uint32_t*} pointer to runner to check
    @returns    {double}    whether or not it is running
 */
GMX double  runnerGetRunning(uint32_t* ptr)
{
    BIC_Runner* runner = (BIC_Runner*) ptr[0];
    return runner -> getRunning();
}

/*
    @desc   Adds a task to the runner to compute.
    @param  {uint32_t*} buffer containing runner and container pointers
    @returns    {void}
 */
GMX void    runnerPushTask(uint32_t* ptr)
{
    BIC_Runner* runner = (BIC_Runner*) ptr[0];
    BIC_Container* task = (BIC_Container*) ptr[1];
    runner -> pushTask(task);
}

/*
    @desc   Attempts to pull a task from the runner. If the runner is
            already computing then it will be stopped and restarted.
            NOTE:   There is no safety in the case of the thread failing
                    to restart!
    @param  {uint32_t*} buffer containing runner ant container pointers
    @returns    {void}
 */
GMX void    runnerPullTask(uint32_t* ptr)
{
    BIC_Runner* runner = (BIC_Runner*) ptr[0];
    BIC_Container* task = (BIC_Container*) ptr[1];

        // If running, attempt to kill and wait.
        // Pull task and restart.
    if (runner -> getRunning())
    {
        runner -> stopThread();
        runner -> joinThread();
        runner -> pullTask(task);

        if (runner -> getTaskCount() > 0)
            runner -> startThread();
    }
    else
        runner -> pullTask(task);
}

/*
    @desc   Sets the value to our GameMaker function called by excall.
    @param  {uint32_t*} pointer to our runner and (in the case of a pointer
                        argument) the length of our buffer
    @param  {uint32_t*} type of data our value is
    @param  {void*}     pointer to our data
    @returns    {void}
 */
GMX void    runnerSetWaitPort(uint32_t* ptr, double type, void* value)
{
    BIC_Runner* runner = (BIC_Runner*) ptr[0];

    port_t port;
    if (type == PT_float)
        port = new double(((double*) value)[0]);

    if (type == PT_int)
        port = new int(((int*) value)[0]);

    if (type == PT_ptr)
    {
        uint8_t* ptr_ = new uint8_t[ptr[1]];
        for (uint32_t i = 0; i < ptr[1]; ++i)
            ptr_[i] = ((uint8_t*)value)[i];

        port = ptr_;
        port.length = ptr[1];
    }

    runner -> setWaitValue(port);
}
#endif
/// -- GameMaker interface definition -- /// END


/// -- ERROR REPORTING -- /// START
void errorCommit(const char * err)
{
    #ifdef BIC_INTERFACE_GAMEMAKER
        gmConnect.lock();
        runner_errors.push_back(err);

        // Send error to GameMaker:
        int mapID;
        mapID = GMMapCreate(0);
        GMMapAddString(mapID, "error", err);
        GMAsyncTrigger(mapID, GM_EVENT);
        gmConnect.unlock();
    #endif
};

void messageCommit(const char * err)
{
    #ifdef BIC_INTERFACE_GAMEMAKER
        gmConnect.lock();
        // Send error to GameMaker:
        int mapID;
        mapID = GMMapCreate(0);
        GMMapAddString(mapID, "message", err);
        GMAsyncTrigger(mapID, GM_EVENT);
        gmConnect.unlock();
    #endif // BIC_INTERFACE_GAMEMAKER
};
/// -- ERROR REPORTING -- /// END

#endif // INTERFACES_H_INCLUDED
